#!/bin/ksh
#
#	$Id: ungrey-robins.ksh,v 1.38 2019/12/18 16:30:39 craski-shell_86 Exp $
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# Copyright (c) 2013, 2015-2017, 2019 Craig R. Skinner <Skinner@Britvault.Co.UK>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#


while getopts "dg:" opt
do
	case ${opt} in
		d)
			DEBUG='aye'
			;;
		g)
			grace="${OPTARG}"
			printf %d ${grace} >/dev/null || exit
			;;
	esac
done


[[ -n ${DEBUG} ]] && set -o xtrace
this=${0##*/}
domains="/etc/mail/${this}.ptrn"


alias logger="logger -t ${this}"


function syslog_listings
{
	[[ -n ${DEBUG} && $# -gt 2 ]] || return
	local ip=${1}
	shift
	local domain=${1}
	shift
	local listings=${*}
	logger -p mail.debug "${ip} listed by domain ${domain} as ${listings}"
}


function tidy_exit
{
	rm ${spamds} ${monos} ${helos} ${robins}
	exit ${1}
}


# Ensure domains list exists, is readable & has dotted domains:
grep -q '^\.' ${domains} ||
{
	error=$?
	print -u2 "Invalid ${domains} file."
	exit ${error}
}


spamds=$(mktemp) || exit
spamdb | fgrep -v 'SPAMTRAP|' > ${spamds} || tidy_exit $?


helos=$(mktemp)
fgrep -f ${domains} ${spamds} | awk -F \| '/^GREY/ { print $2, $3 }' > ${helos}
[[ -s ${helos} ]] || tidy_exit


# Round robin SMTP clients for whitelisting via spamdb
robins=$(mktemp)


# Traps are blacklisted, black & white = monochrome
monos=$(mktemp)
awk -F \| '/^WHITE|^TRAPPED/ { print $2 }' ${spamds} > ${monos}


# Skip IPs already whitelisted or trapped:
fgrep -v -f ${monos} ${helos} |
while read ip helo
do
	# Insist there is a DNS PTR record, and ...
	ptr=$(dig -x ${ip} +short) ||
	{
		logger -p mail.warning "${ip} skipped due to no DNS PTR record."
		continue
	}

	# ... either the IP's DNS PTR record matches the HELO hostname:
	[[ ${ptr} == "${helo}." ]] ||
	{
		# ... or the IP's DNS PTR record matches the IP address:
		[[ $(dig ${ptr} +short) == ${ip} ]] ||
		{
			logger -p mail.warning "${ip} skipped as DNS PTR doesn't match IP nor SMTP HELO."
			continue
		}
	}

	# Require the IP to be DNS whitelisted:
	unset score
	set -A octets $(print ${ip} | tr -s '.' ' ')
	rev_ip="${octets[3]}.${octets[2]}.${octets[1]}.${octets[0]}"

	# http://www.dnswl.org/?page_id=15
	listing=$(dig ${rev_ip}.list.dnswl.org +short)
	syslog_listings ${ip} list.dnswl.org ${listing}
	case ${listing} in
		127.0.*.0)
			# none - only avoid outright blocking (e.g: Hotmail, Yahoo, etc. mailservers)
			let score+=1
			;;
		127.0.*.1)
			# low - reduce chance of false positives
			let score+=2
			;;
		127.0.*.2)
			# medium - make sure to avoid false positives but allow override for clear cases
			let score+=5
			;;
		127.0.*.3)
			# high - avoid override
			let score+=7
			;;
	esac

	# http://www.mailspike.net
	listing=$(dig ${rev_ip}.rep.mailspike.net +short)
	syslog_listings ${ip} rep.mailspike.net ${listing}
	case ${listing} in
		127.0.0.10)
			# Worst possible reputation
			let score-=7
			;;
		127.0.0.11)
			# Very bad reputation
			let score-=6
			;;
		127.0.0.12)
			# Bad reputation
			let score-=5
			;;
		127.0.0.13)
			# Suspicious behaviour
			let score-=4
			;;
		127.0.0.14)
			# Neutral, probably spam
			let score-=3
			;;
		127.0.0.16)
			# Neutral, probably legitimate
			let score+=3
			;;
		127.0.0.17)
			# Possible legitimate sender
			let score+=4
			;;
		127.0.0.18)
			# Good reputation
			let score+=5
			;;
		127.0.0.19)
			# Very good reputation
			let score+=6
			;;
		127.0.0.20)
			# Excellent reputation
			let score+=7
			;;
	esac

	# http://www.senderscore.com
	listing=$(dig ${rev_ip}.score.senderscore.com +short)
	syslog_listings ${ip} score.senderscore.com ${listing}
	case ${listing} in
		127.0.*.[0-9])
			let score-=7
			;;
		127.0.*.1[0-9])
			let score-=6
			;;
		127.0.*.2[0-9])
			let score-=5
			;;
		127.0.*.3[0-9])
			let score-=4
			;;
		127.0.*.4[0-5])
			let score-=3
			;;
		127.0.*.5[5-9])
			let score+=3
			;;
		127.0.*.6[0-9])
			let score+=4
			;;
		127.0.*.7[0-9])
			let score+=5
			;;
		127.0.*.8[0-9])
			let score+=6
			;;
		127.0.*.9[0-9])
			let score+=7
			;;
	esac

	# http://hostkarma.junkemailfilter.com
	set -A listings $(dig ${rev_ip}.hostkarma.junkemailfilter.com +short)
	syslog_listings ${ip} hostkarma.junkemailfilter.com ${listings[@]}
	for listing in ${listings[@]}
	do
		case ${listing} in
			# http://wiki.junkemailfilter.com/index.php/Spam_DNS_Lists#How_to_use_the_Lists
			127.0.0.1)
				# whilelist - trusted nonspam
				let score+=7
				;;
			127.0.0.2)
				# blacklist - block spam
				let score-=7
				;;
			127.0.0.3)
				# yellowlist - mix of spam and nonspam
				let score+=3
				;;
			127.0.0.4)
				# brownlist - all spam, but not yet enough to blacklist
				let score-=5
				;;
			127.0.0.5)
				# NOBL - not a spam only source and no blacklists need to be tested
				let score+=5
				;;
			# http://wiki.junkemailfilter.com/index.php/Spam_DNS_Lists#Tracking_use_of_QUIT
			127.0.1.1)
				# QUIT is used
				let score+=1
				;;
			127.0.1.2)
				# No QUIT is used
				let score-=1
				;;
		esac
	done

	# Default grace threshold: 7
	[[ ${score} -ge ${grace:-7} ]] &&
	{
		print ${ip} >> ${robins}
		score="${score} (granted grace)"
	}
	logger -p mail.info "${ip} ${ptr} scored ${score}"
done


[[ -s ${robins} ]] || tidy_exit
spamdb -a $(< ${robins}) || exit
logger -p mail.notice "spamdb whitelisted: $(< ${robins})" && tidy_exit

